import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_lower_interface.nusmv.dd import dd as nsdd

from pynusmv_tools.mas import glob

from pynusmv_tools.atlkPO.check import check
from pynusmv_tools.atlkFO.parsing import parseATLK


class TestCheck(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
    
    
    def little(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/little.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame_post_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame-post-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def trans2_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/2-transmission-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/transmission.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission_with_knowledge(self):
        glob.load_from_file(
                        "tests/pynusmv_tools/atlkPO/models/transmission-knowledge.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/transmission-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
    
    def transmission_post_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/transmission-post-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def show_si(self, fsm, bdd):
        if bdd.isnot_false():
            sis = fsm.pick_all_states_inputs(bdd)
            print("SI count:", len(sis))
            for si in sis:
                print(si.get_str_values())
            
    def show_s(self, fsm, bdd):
        if bdd.isnot_false():
            for s in fsm.pick_all_states(bdd):
                print(s.get_str_values())
            
    def show_i(self, fsm, bdd):
        if bdd.isnot_false():
            for i in fsm.pick_all_inputs(bdd):
                print(i.get_str_values())
        
        
    def test_cardgame_not_improved(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0]))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0]))
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0]))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0]))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0]))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0]))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0]))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0]))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0]))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0]))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0]))
        
    
    @unittest.expectedFailure # MC algo does not deal with fairness on action for now
    def test_cardgame_fair_not_improved(self):
        fsm = self.cardgame_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0]))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0]))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0]))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0]))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0]))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0]))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0]))
        
    
    def test_cardgame_post_fair_not_improved(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0]))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0]))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0]))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0]))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0]))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0]))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0]))
        

    def test_transmission_not_improved(self):
        fsm = self.transmission()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0]))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0]))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0]))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0]))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0]))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0]))
        
    
    def test_transmission_with_know_not_improved(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0]))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0]))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0]))
    
    
    @unittest.expectedFailure # MC algo does not deal with fairness on action for now
    def test_transmission_fair_not_improved(self):
        fsm = self.transmission_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0]))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0]))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0]))
    
    
    def test_transmission_post_fair_not_improved(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0]))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0]))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0]))
        
        
    def test_transmission_improved(self):
        fsm = self.transmission()
        
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="SF"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="SF"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="SF"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="SF"))
        
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="SF"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="SF"))
        
    
    def test_transmission_with_know_improved(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="SF"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="SF"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="SF"))
        
    
    def test_transmission_post_fair_improved(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="SF"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="SF"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="SF"))
        
        
    def test_cardgame_improved(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="SF"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="SF"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="SF"))
        
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="SF"))
        
        
    @unittest.skip("Model checking takes too long.")
    def test_cardgame_post_fair_improved(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], variant="SF"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="SF"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="SF"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="SF"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="SF"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], variant="SF"))
        
    
    def test_transmission_FSF(self):
        fsm = self.transmission()
        
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FSF"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FSF"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FSF"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FSF"))
        
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FSF"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FSF"))
        
    def test_transmission_with_know_FSF(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FSF"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FSF"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FSF"))
        
    def test_transmission_post_fair_FSF(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FSF"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FSF"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FSF"))
        
        
    def test_cardgame_FSF(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="FSF"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="FSF"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="FSF"))
        
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="FSF"))
        
        
    @unittest.skip("Model checking takes too long.")
    def test_cardgame_post_fair_FSF(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], variant="FSF"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="FSF"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="FSF"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="FSF"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="FSF"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], variant="FSF"))
        

    def test_cardgame_not_improved_gen(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="generator"))
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="generator"))
        
    
    @unittest.expectedFailure # MC algo does not deal with fairness on action for now
    def test_cardgame_fair_not_improved_gen(self):
        fsm = self.cardgame_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="generator"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="generator"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="generator"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="generator"))
        
    
    def test_cardgame_post_fair_not_improved_gen(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="generator"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="generator"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="generator"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="generator"))
        

    def test_transmission_not_improved_gen(self):
        fsm = self.transmission()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="generator"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="generator"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="generator"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="generator"))
        
    
    def test_transmission_with_know_not_improved_gen(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="generator"))
    
    
    @unittest.expectedFailure # MC algo does not deal with fairness on action for now
    def test_transmission_fair_not_improved_gen(self):
        fsm = self.transmission_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="generator"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="generator"))
    
    
    def test_transmission_post_fair_not_improved_gen(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="generator"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="generator"))
        
        
    def test_transmission_improved_gen(self):
        fsm = self.transmission()
        
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="SF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="SF", implem="generator"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="SF", implem="generator"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="SF", implem="generator"))
        
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="SF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="SF", implem="generator"))
        
    
    def test_transmission_with_know_improved_gen(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="SF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="SF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="SF", implem="generator"))
        
    
    def test_transmission_post_fair_improved_gen(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="SF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="SF", implem="generator"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="SF", implem="generator"))
        
        
    def test_cardgame_improved_gen(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="SF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="SF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="SF", implem="generator"))
        
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="SF", implem="generator"))
        
        
    @unittest.skip("Model checking takes too long.")
    def test_cardgame_post_fair_improved_gen(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], variant="SF", implem="generator"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="SF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="SF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="SF", implem="generator"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="SF", implem="generator"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], variant="SF", implem="generator"))
        
    
    def test_transmission_FSF_gen(self):
        fsm = self.transmission()
        
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FSF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FSF", implem="generator"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FSF", implem="generator"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FSF", implem="generator"))
        
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FSF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FSF", implem="generator"))
        
    def test_transmission_with_know_FSF_gen(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FSF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FSF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FSF", implem="generator"))
        
    def test_transmission_post_fair_FSF_gen(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FSF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FSF", implem="generator"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FSF", implem="generator"))
        
        
    def test_cardgame_FSF_gen(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="FSF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="FSF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="FSF", implem="generator"))
        
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="FSF", implem="generator"))
        
        
    @unittest.skip("Model checking takes too long.")
    def test_cardgame_post_fair_FSF_gen(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], variant="FSF", implem="generator"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="FSF", implem="generator"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="FSF", implem="generator"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="FSF", implem="generator"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="FSF", implem="generator"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], variant="FSF", implem="generator"))
        
        
        
    def test_cardgame_not_improved_opt(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="optimized"))
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="optimized"))
        
    
    @unittest.expectedFailure # MC algo does not deal with fairness on action for now
    def test_cardgame_fair_not_improved_opt(self):
        fsm = self.cardgame_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="optimized"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="optimized"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="optimized"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="optimized"))
        
    
    def test_cardgame_post_fair_not_improved_opt(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="optimized"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="optimized"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="optimized"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="optimized"))
        

    def test_transmission_not_improved_opt(self):
        fsm = self.transmission()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="optimized"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="optimized"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="optimized"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="optimized"))
        
    
    def test_transmission_with_know_not_improved_opt(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="optimized"))
    
    
    @unittest.expectedFailure # MC algo does not deal with fairness on action for now
    def test_transmission_fair_not_improved_opt(self):
        fsm = self.transmission_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="optimized"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="optimized"))
    
    
    def test_transmission_post_fair_not_improved_opt(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="optimized"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="optimized"))
        
    
    def test_transmission_FSF_opt(self):
        fsm = self.transmission()
        
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FSF", implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FSF", implem="optimized"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FSF", implem="optimized"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FSF", implem="optimized"))
        
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FSF", implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FSF", implem="optimized"))
        
    def test_transmission_with_know_FSF_opt(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FSF", implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FSF", implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FSF", implem="optimized"))
        
    def test_transmission_post_fair_FSF_opt(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FSF", implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FSF", implem="optimized"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FSF", implem="optimized"))
        
        
    def test_cardgame_FSF_opt(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="FSF", implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="FSF", implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="FSF", implem="optimized"))
        
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="FSF", implem="optimized"))
        
        
    @unittest.skip("Model checking takes too long.")
    def test_cardgame_post_fair_FSF_opt(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], variant="FSF", implem="optimized"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="FSF", implem="optimized"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="FSF", implem="optimized"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="FSF", implem="optimized"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="FSF", implem="optimized"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], variant="FSF", implem="optimized"))
        
    
    def test_cardgame_not_improved_mem(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="memory"))
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="memory"))
        
    
    @unittest.expectedFailure # MC algo does not deal with fairness on action for now
    def test_cardgame_fair_not_improved_mem(self):
        fsm = self.cardgame_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="memory"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="memory"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="memory"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="memory"))
        
    
    def test_cardgame_post_fair_not_improved_mem(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="memory"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="memory"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="memory"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="memory"))
        

    def test_transmission_not_improved_mem(self):
        fsm = self.transmission()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="memory"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="memory"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="memory"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="memory"))
        
    
    def test_transmission_with_know_not_improved_mem(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="memory"))
    
    
    @unittest.expectedFailure # MC algo does not deal with fairness on action for now
    def test_transmission_fair_not_improved_mem(self):
        fsm = self.transmission_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="memory"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="memory"))
    
    
    def test_transmission_post_fair_not_improved_mem(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="memory"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="memory"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="memory"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="memory"))


#! /usr/local/bin/python3
"""
This is an alternative implementation which uses the native C api as they were
brought by the the ltlspec module.

Concretely, it uses `ltlspec.generate_ltl_problem` to generate the complete
problem for an LTL verification. (This api fully originates from nusmv).
"""
import sys
import argparse

from pynusmv.init         import init_nusmv
from pynusmv.glob         import load
from pynusmv.bmc.glob     import BmcSupport, master_be_fsm
from pynusmv.parser       import parse_ltl_spec
from pynusmv.node         import Node
from pynusmv.sat          import SatSolverFactory, Polarity, SatSolverResult
from pynusmv.bmc          import ltlspec, utils as bmcutils

def arguments():
    """
    Creates the arguments parser and manages to react to wrong usage.

    :returns: an object having field to store each of the command line arguments
    """
    parser = argparse.ArgumentParser(description="a PyNuSMV backed LTL sat based bmc verifier for LTL")
    parser.add_argument("-v", "--verbose", action="store_true", help="Displays the text of the analyzed model")
    parser.add_argument("-k", "--bound",   type=int, default=10, help="the maximum number of steps in a verified path")
    parser.add_argument("-s", "--spec",    type=str, help="the LTL specification to verify")
    parser.add_argument("-f", "--no-fairness",  help="disable the use of fairness constraints", action="store_true")
    parser.add_argument("-i", "--no-invariants",help="disable the invariants enforcement", action="store_true")
    parser.add_argument("-d", "--dry-run", action="store_true", help="do not perform the verification (no sat solving)")
    parser.add_argument("model", type=str, help="the name of a file containing an SMV model")

    return parser.parse_args()

def check_problem(pb, length):
    fsm    = master_be_fsm()
    cnf    = pb.to_cnf(Polarity.POSITIVE)

    solver = SatSolverFactory.create()
    solver+= cnf
    solver.polarity(cnf, Polarity.POSITIVE)

    if solver.solve() == SatSolverResult.SATISFIABLE:
        cnt_ex = bmcutils.generate_counter_example(fsm, pb, solver, length, "Violation")
        return ("Violation", cnt_ex)
    else:
        return ("Ok", None)

def check_ltl(fml, bound, dry_run):
    import time
    fsm     = master_be_fsm()

    for i in range(bound+1):
        start = time.time()
        problem = ltlspec.generate_ltl_problem(fsm, fml, i)
        end   = time.time()
        if not dry_run:
            status, trace = check_problem(problem, i)
            if status != "Ok":
                return (status, i, trace)
            else:
                print("-- No problem at length {}".format(i))
        else:
            print(" 'Problem {}' ; {}".format(i, end-start))

    return ("Ok", bound, None)

def check(formula, args):
    parsed_fml          = Node.from_ptr(parse_ltl_spec(formula.strip()))
    status,length,trace = check_ltl(parsed_fml, args.bound, args.dry_run)
    if status != 'Ok':
        print("-- {} for length {}".format(status, length))
        print(trace)


def main():
    args = arguments()

    with init_nusmv():
        load(args.model)
        if args.verbose:
            with open(args.model) as f:
                print(f.read())

        with BmcSupport():
            if args.spec is not None:
                check(args.spec, args)
            else:
                print("Enter LTL properties, one per line:")
                for line in sys.stdin:
                    check(line, args)

if __name__ == "__main__":
    main()



"""
AST classes for ATL formulas representation.
"""

from collections import namedtuple

class Spec:
    """A Spec is represents a ATL formula."""

# TrueExp, FalseExp, Init, Reachable
TrueExp = namedtuple('TrueExp', [])
TrueExp.__bases__ += (Spec,)
TrueExp.__str__ = lambda self: "True"

FalseExp = namedtuple('FalseExp', [])
FalseExp.__bases__ += (Spec,)
FalseExp.__str__ = lambda self: "False"

# Atom
Atom = namedtuple('Atom', ['value'])
Atom.__bases__ += (Spec,)
Atom.__str__ = lambda self: "'" + str(self.value) + "'"

# logical : Not, And, Or, Implies, Iff
Not = namedtuple('Not', ['child'])
Not.__bases__ += (Spec,)
Not.__str__ = lambda self: "~" + "(" + str(self.child) + ")"

And = namedtuple('And', ['left', 'right'])
And.__bases__ += (Spec,)
And.__str__ = lambda self: "(" + str(self.left) + " & " + str(self.right) + ")"

Or = namedtuple('Or', ['left', 'right'])
Or.__bases__ += (Spec,)
Or.__str__ = lambda self: "(" + str(self.left) + " | " + str(self.right) + ")"

Implies = namedtuple('Implies', ['left', 'right'])
Implies.__bases__ += (Spec,)
Implies.__str__ = (lambda self: "(" + str(self.left) + " -> " +
                                      str(self.right) + ")")

Iff = namedtuple('Iff', ['left', 'right'])
Iff.__bases__ += (Spec,)
Iff.__str__ = (lambda self: "(" + str(self.left) + " <-> " +
                                  str(self.right) + ")")
                                  

# strategic : <g>F, <g>G, <g>X, <g>U, <g>W, [g]F, [g]G, [g]X, [g]U, [g]W
CEF = namedtuple('CEF', ['group', 'child'])
CEF.__bases__ += (Spec,)
CEF.__str__ = (lambda self: "<" +','.join([ag.value for ag in self.group])+ ">"+
                            "F " + str(self.child))
                            
CEG = namedtuple('CEG', ['group', 'child'])
CEG.__bases__ += (Spec,)
CEG.__str__ = (lambda self: "<" +','.join([ag.value for ag in self.group])+ ">"+
                            "G " + str(self.child))
                           
CEX = namedtuple('CEX', ['group', 'child'])
CEX.__bases__ += (Spec,)
CEX.__str__ = (lambda self: "<" +','.join([ag.value for ag in self.group])+ ">"+ 
                            "X " + str(self.child))
                           
CEU = namedtuple('CEU', ['group', 'left', 'right'])
CEU.__bases__ += (Spec,)
CEU.__str__ = (lambda self: "<" +','.join([ag.value for ag in self.group])+ ">"+ 
                            "[" + str(self.left) +
                            " U " + str(self.right) + "]")
                           
CEW = namedtuple('CEW', ['group', 'left', 'right'])
CEW.__bases__ += (Spec,)
CEW.__str__ = (lambda self: "<" +','.join([ag.value for ag in self.group])+ ">"+ 
                            "[" + str(self.left) +
                            " W " + str(self.right) + "]")
 
CAF = namedtuple('CAF', ['group', 'child'])
CAF.__bases__ += (Spec,)
CAF.__str__ = (lambda self: "[" +','.join([ag.value for ag in self.group])+ "]"+ 
                            "F " + str(self.child))
                           
CAG = namedtuple('CAG', ['group', 'child'])
CAG.__bases__ += (Spec,)
CAG.__str__ = (lambda self: "[" +','.join([ag.value for ag in self.group])+ "]"+ 
                            "G " + str(self.child))
                           
CAX = namedtuple('CAX', ['group', 'child'])
CAX.__bases__ += (Spec,)
CAX.__str__ = (lambda self: "[" +','.join([ag.value for ag in self.group])+ "]"+ 
                            "X " + str(self.child))
                           
CAU = namedtuple('CAU', ['group', 'left', 'right'])
CAU.__bases__ += (Spec,)
CAU.__str__ = (lambda self: "[" +','.join([ag.value for ag in self.group])+ "]"+ 
                            "[" + str(self.left) +
                            " U " + str(self.right) + "]")
                           
CAW = namedtuple('CAW', ['group', 'left', 'right'])
CAW.__bases__ += (Spec,)
CAW.__str__ = (lambda self: "[" +','.join([ag.value for ag in self.group])+ "]"+ 
                            "[" + str(self.left) +
                            " W " + str(self.right) + "]")


import unittest

from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv import glob

from pynusmv_lower_interface.nusmv.parser import parser as nsparser
from pynusmv_lower_interface.nusmv.compile import compile as nscompile
from pynusmv_lower_interface.nusmv.node import node as nsnode
from pynusmv_lower_interface.nusmv.set import set as nsset
from pynusmv_lower_interface.nusmv.utils import utils as nsutils
from pynusmv_lower_interface.nusmv.compile.symb_table import symb_table as nssymb_table

class TestVariables(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        
    def tearDown(self):
        deinit_nusmv()
        
    def test_fsm(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/dining-crypto.smv")
        glob.compute_model()
        fsm = glob.prop_database().master.bddFsm
        self.assertIsNotNone(fsm)
        
        
    def test_printing_parsed_tree(self):
        car = nsnode.car
        cdr = nsnode.cdr
        
        glob.load_from_file("tests/pynusmv_tools/ctlk/dining-crypto.smv")
        
        # Get parsed tree
        tree = nsparser.cvar.parsed_tree
        
        
        print(tree.type) # 145 = CONS
        print(car(tree).type) # 117 = MODULE
        print(car(car(tree)).type) # 119 = MODTYPE
        print(nsnode.sprint_node(car(car(car(tree))))) # main
        print(cdr(tree).type) # 145 = CONS
        print(car(cdr(tree)).type) # 117 = MODULE
        print(car(car(car(cdr(tree)))).type) # 161 = ATOM
        print(nsnode.sprint_node(car(car(car(cdr(tree)))))) # cryptograph
        print(cdr(cdr(tree))) # None
        
        
        
        
    def get_instances_for_module(self, modtree):
        """
        Return the list of variables that are instances of modules
        in module modtree.
        
        modtree is a part of the AST of the SMV model. modtree.type = MODULE
        """
        # MODULE(MODTYPE(...), declarationlist)
        varlist = []
        declarations = nsnode.cdr(modtree)
        while declarations is not None:
            decl = nsnode.car(declarations)
            if decl.type == nsparser.VAR:
                decl = nsnode.car(decl)
                while decl is not None:
                    var = nsnode.car(decl)
                    
                    # var_id : type => COLON(ATOM, type)
                    if nsnode.cdr(var).type == nsparser.MODTYPE:
                        varlist.append(var)
                    
                    decl = nsnode.cdr(decl)
                            
            declarations = nsnode.cdr(declarations)
            
        return varlist
        
        
    def print_instances(self, filepath):
        print("----- Instances for", filepath)
        
        car = nsnode.car
        cdr = nsnode.cdr
        
        glob.load_from_file(filepath)
        
        # Get parsed tree
        tree = nsparser.cvar.parsed_tree
        
        
        print("--- Main instances")
        # main module is car(tree)
        main_vars = self.get_instances_for_module(car(tree))
        
        instances_args = {}
        
        for var in main_vars:
            # var = COLON(ATOM, MODTYPE(ATOM, CONS))
            
            varname = nsnode.sprint_node(car(var))
            instances_args[varname] = []
            
            args = cdr(cdr(var))
            argslist = []
            while args is not None:
                arg = car(args)
                instances_args[varname].append(arg)
                argslist.append(nsnode.sprint_node(arg))
                args = cdr(args)
                
            print(varname, ":", nsnode.sprint_node(car(cdr(var))), argslist)
                
        
        # Get FSM and stuff
        glob.compute_model()
        fsm = glob.prop_database().master.bddFsm
        self.assertIsNotNone(fsm)
        
        flatH = nscompile.cvar.mainFlatHierarchy
        st = nscompile.Compile_get_global_symb_table()
        self.assertIsNotNone(flatH)
        
        
        print("--- Check arguments instances")
        for instance in instances_args:
            print("INSTANCE", instance)
            for arg in instances_args[instance]:                
                arg, err = nscompile.FlattenSexp(st, arg, None)
                self.assertEqual(err, 0)
                isVar = nssymb_table.SymbTable_is_symbol_var(st, arg)
                if isVar:
                    print("VAR", nsnode.sprint_node(arg))
                else:
                    print("NOT VAR", nsnode.sprint_node(arg))
        
        
        print("--- All vars")
        varset = nscompile.FlatHierarchy_get_vars(flatH)
        self.assertIsNotNone(varset)
        varlist = nsset.Set_Set2List(varset)
        self.assertIsNotNone(varlist)
        
        ite = nsutils.NodeList_get_first_iter(varlist)
        while not nsutils.ListIter_is_end(ite):
            var = nsutils.NodeList_get_elem_at(varlist, ite)
            isInput = nssymb_table.SymbTable_is_symbol_input_var(st, var)
            isVar = nssymb_table.SymbTable_is_symbol_var(st, var)
            if isInput:
                print("IVAR", "\t",
                      "IN", "'" + nsnode.sprint_node(car(var)) + "'", "\t",
                      nsnode.sprint_node(var))
            elif isVar:
                print("VAR", "\t",
                      "IN", "'" + nsnode.sprint_node(car(var)) + "'", "\t",
                      nsnode.sprint_node(var))
            else:
                print("[ERROR] Unknown type:", nsnode.sprint_node(var))
            ite = nsutils.ListIter_get_next(ite)
            
        print("------------------------------------------------------")
        
        
    def test_get_instances_after_flattening(self):
        car = nsnode.car
        cdr = nsnode.cdr
        
        glob.load_from_file("tests/pynusmv_tools/ctlk/dining-crypto.smv")
        
        # Flatten
        glob.flatten_hierarchy()
        
        # Get parsed tree
        tree = nsparser.cvar.parsed_tree
        self.assertIsNotNone(tree)
        self.assertIsNotNone(car(tree))
        
        print(tree.type) # 145 = CONS
        print(car(tree).type) # 117 = MODULE
        print(nsnode.sprint_node(car(car(car(tree))))) # main
        print(cdr(car(tree)).type) # 145 = CONS
        print(car(cdr(car(tree))).type) # 113 = DEFINE
        print(cdr(cdr(car(tree)))) # None 
        print(cdr(tree).type) # 145 = CONS
        print(car(cdr(tree)).type) # 117 = MODULE
        print(nsnode.sprint_node(car(car(car(cdr(tree)))))) # cryptograph
        
        print("----- Instances after flattening")
        # main module is car(tree)
        main_vars = self.get_instances_for_module(car(tree))
        # Variables are removed from parsed_tree when flattening
        self.assertEqual(len(main_vars), 0)
        
        instances_args = {}
        
        for var in main_vars:
            # var = COLON(ATOM, MODTYPE(ATOM, CONS))
            
            varname = nsnode.sprint_node(car(var))
            instances_args[varname] = []
            
            args = cdr(cdr(var))
            argslist = []
            while args is not None:
                arg = car(args)
                instances_args[varname].append(arg)
                argslist.append(nsnode.sprint_node(arg))
                args = cdr(args)
                
            print(varname, ":", nsnode.sprint_node(car(cdr(var))), argslist)
            
        print("------------------------------------------------------")

            
    def test_dincry(self):
        self.print_instances("tests/pynusmv_tools/ctlk/dining-crypto.smv")
            
        
    def test_bitCounter(self):
        self.print_instances("tests/pynusmv_tools/ctlk/bitCounter.smv")
       
        
    def test_counters(self):
        self.print_instances("tests/pynusmv_tools/ctlk/counters.smv")


import unittest

from pynusmv.fsm import BddFsm
from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.parser import parse_simple_expression as parseSexp
from pynusmv import prop
from pynusmv_lower_interface.nusmv.opt import opt as nsopt

from pynusmv_tools.ctl.eval import eval_ctl


class TestEval(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        deinit_nusmv()
    
    
    def init_model(self):
        fsm = BddFsm.from_filename("tests/pynusmv_tools/ctl/admin.smv")
        self.assertIsNotNone(fsm)
        return fsm
    
    
    def nameche_model(self):
        fsm = BddFsm.from_filename("tests/pynusmv_tools/ctl/nameche08R1.smv")
        self.assertIsNotNone(fsm)
        return fsm
    

    def test_atom(self):
        fsm = self.init_model()
        
        anspec = prop.Spec(parseSexp("admin = none"))
        an = eval_ctl(fsm, anspec)
        self.assertIsNotNone(an)
        self.assertTrue(fsm.init <= an)
        
        aaspec = prop.Spec(parseSexp("admin = alice"))
        aa = eval_ctl(fsm, aaspec)
        self.assertIsNotNone(aa)
        self.assertTrue((fsm.init & aa).is_false())
        self.assertTrue((aa & an).is_false())
    
    def test_ctl(self):
        fsm = self.init_model()
        # EF admin = alice is True
        efaa = prop.ef(prop.atom("admin = alice"))
        self.assertTrue(fsm.init <= eval_ctl(fsm, efaa))
        
        # EG admin != alice is True
        egana = prop.eg(prop.atom("admin != alice"))
        self.assertTrue(fsm.init <= eval_ctl(fsm, egana))
        
        # AF admin != none is True
        afann = prop.af(prop.atom("admin != none"))
        self.assertTrue(fsm.init <= eval_ctl(fsm, afann))
        
        # AG admin in {alice, bob, none} is True
        aga = prop.ag(prop.atom("admin in {alice, bob, none}"))
        self.assertTrue(fsm.init <= eval_ctl(fsm, aga))
        
        # AG (admin != none -> AG admin != none) is True
        aganniann = prop.ag(prop.imply(prop.atom("admin != none"),
                                       prop.ag(prop.atom("admin != none"))))
        self.assertTrue(fsm.init <= eval_ctl(fsm, aganniann))
        
        # AG (admin = alice -> AX admin = alice) is True
        agaaiaxaa = prop.ag(prop.imply(prop.atom("admin = alice"),
                                       prop.ax(prop.atom("admin = alice"))))
        self.assertTrue(fsm.init <= eval_ctl(fsm, agaaiaxaa))
        
        
        # AX admin = alice is False
        axaa = prop.ax(prop.atom("admin = alice"))
        self.assertFalse(fsm.init <= eval_ctl(fsm, axaa))
        
        # EX admin != none is False
        exann = prop.ex(prop.atom("admin != none"))
        self.assertFalse(fsm.init <= eval_ctl(fsm, exann))
        
        # AG admin = none is False
        agan = prop.ag(prop.atom("admin = none"))
        self.assertFalse(fsm.init <= eval_ctl(fsm, agan))
        
        # EG admin = none is False
        egan = prop.eg(prop.atom("admin = none"))
        self.assertFalse(fsm.init <= eval_ctl(fsm, egan))
    
#    @unittest.skip("Takes too long")
    def test_namecheR1(self):
        # activate bdd dynamic reordering
        nsopt.set_dynamic_reorder(nsopt.OptsHandler_get_instance())
        
        fsm = self.nameche_model()
        
        # AG (T_04BM.st = o & TRP_CM.krc = s -> AX (!R1.L_CS))
        to = prop.atom("T_04BM.st = o")
        trps = prop.atom("TRP_CM.krc = s")
        nr1lcs = prop.atom("!R1.L_CS")
        p = prop.imply(to & trps, prop.ax(nr1lcs))
        self.assertTrue(fsm.reachable_states <= eval_ctl(fsm, p))
        
        # AG (EF R1.L_CS)
        r1lcs = prop.atom("R1.L_CS")
        p = prop.ef(r1lcs)
        self.assertTrue(fsm.reachable_states <= eval_ctl(fsm, p))


'''
# This script provides all the necessary logic for you to build a distribution
# of pynusmv-tools which you can run on your machine.
#
# This file is part of the pynusmv-tools distribution. As such it is licensed to you
# under the term of the LGPLv2. For more information regarding the legal aspect
# of this licensing, please refer to the full text of the license on the free
# software foundation website.
#
# Author: X. Gillard <xavier.gillard [at] uclouvain.be>
'''
from setuptools import setup, find_packages

setup(name             = 'pynusmv-tools',
      version          = "1.0rc6",
      author           = "Simon BUSARD, Xavier GILLARD",
      author_email     = "simon.busard@uclouvain.be, xavier.gillard@uclouvain.be",
      url              = "http://lvl.info.ucl.ac.be/Tools/PyNuSMV",
      description      = "Tools, examples and experiments that showcase the potential uses for PyNuSMV",
      packages         = find_packages(),
      install_requires = ['pynusmv'],
      entry_points     = {
        'console_scripts' : [
            'ctl=pynusmv_tools.ctl.CTLcheck:main',
            'fairctl=pynusmv_tools.fairctl.FairCTLcheck:main',
            'ctlk=pynusmv_tools.ctlk.cmd.cmd:main',
            'arctl=pynusmv_tools.arctl.cmd.trace:main',
            'tlace=pynusmv_tools.tlace.tlace:main',
            # MAS
            'atl=pynusmv_tools.atl.check:main',
            'atlk_fo=pynusmv_tools.atlkFO.check:main',
            'atlk_po=pynusmv_tools.atlkPO.check:main',
            # utils
            'smv2dot=pynusmv_tools.dotDump:main',
            'smv_cmp=pynusmv_tools.compare:main',
            # BMC tools
            'diagnos=pynusmv_tools.diagnosability:main',
            # LTL BMC - directly using NuSMV api to generate the problem (fastest)
            'bmc_ltl=pynusmv_tools.bmcLTL.bmc_ltl:main',
            # LTL BMC - using the lower interface functions to generate the
            # problem (close to a performance tie with alternative1 + most flexible
            # implementation)
            'bmc_ltl_li=pynusmv_tools.bmcLTL.bmc_ltl_li:main',
            # LTL BMC - not using apis from pynusmv.bmc.*
            'bmc_ltl_py=pynusmv_tools.bmcLTL.bmc_ltl_py:main'
        ]
      },
      # TESTS
      test_suite='nose.collector',
      tests_require=['nose']
)



class LockStack():
    """
    A lock stack is a stack with an additional lock.
    
    When popping, the lock is activated, but the top value is actually removed
    only when the lock is already activated. Whenever pushing a value, the lock
    is deactivated.
    This means that to remove the top, you have to pop twice in a raw
    (without pushing between the two pops).
    The stack can be unlocked.
    """
    
    def __init__(self):
        self.__stack = []
        self.__lock = False
        
    def pop(self):
        if len(self.__stack) <= 0:
            raise IndexError("Empty stack")
        if self.__lock:
            self.__stack.pop()
        else:
            self.__lock = True
            
    def push(self, value):
        self.__stack.append(value)
        self.__lock = False
        
    def top(self):
        if len(self.__stack) <= 0:
            raise IndexError("Empty stack")
        return self.__stack[-1]
        
    def isEmpty(self):
        return len(self.__stack) == 0
        
    def all(self):
        return tuple(self.__stack)
        
    def unlock(self):
        self.__lock = False
        
    def copy(self):
        copy = LockStack()
        for e in self.__stack:
            copy.push(e)
        return copy


import unittest

from pyparsing import ParseException

from pynusmv_tools.atlkFO.parsing import parseATLK
from pynusmv_tools.atlkFO.ast import (Spec, TrueExp, FalseExp, Init, Reachable,
                            Atom, Not, And, Or, Implies, Iff, 
                            AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                            nK, nE, nD, nC, K, E, D, C,
                            CAF, CAG, CAX, CAU, CAW, CEF, CEG, CEX, CEU, CEW)


class TestParsing(unittest.TestCase):
        
    def test_atom(self):
        s = "'c <= 3'"
        asts = parseATLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, Atom))
        self.assertEqual(ast.value, "c <= 3")
        
    
    def test_fail(self):
        kos = ["", "'test", "A<'test'>H 'q'", "O", "A<'a'>E<'e'>G 'true'",
               "E<'ac'>[A<'ac'>['a' E<'ac'>['b' W 'c'] W 'd'] "
               "U A<'ac'>['e' U 'f']]", "A<'t't'>G 'q'"]
        
        for s in kos:
            with self.assertRaises(ParseException):
                parseATLK(s)
                
                
    def test_CEF(self):
        s = "<'agent'>F 'test'"
        asts = parseATLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, CEF))
        for ag in ast.group:
            self.assertTrue(isinstance(ag, Atom))
        self.assertEqual(len(ast.group), 1)
        self.assertEqual(ast.group[0].value, 'agent')
        self.assertTrue(isinstance(ast.child, Atom))
        self.assertEqual(ast.child.value, "test")
        
        
    def test_deep_spec(self):
        s = ("<'a','b'>[['c'][['e'] F ('p' & 'q') W K<'a'> 'p'] U <'d'>G "
             "nC<'g','e'> ('c' -> EF 'k')]")
        asts = parseATLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, CEU))
        
        self.assertEqual(len(ast.group), 2)
        self.assertTrue(isinstance(ast.group[0], Atom))
        self.assertEqual(ast.group[0].value, "a")
        self.assertTrue(isinstance(ast.group[1], Atom))
        self.assertEqual(ast.group[1].value, "b")
        
        self.assertTrue(isinstance(ast.left, CAW))
        self.assertEqual(len(ast.left.group), 1)
        self.assertTrue(isinstance(ast.left.group[0], Atom))
        self.assertEqual(ast.left.group[0].value, "c")
        self.assertTrue(isinstance(ast.left.left, CAF))
        self.assertEqual(len(ast.left.left.group), 1)
        self.assertTrue(isinstance(ast.left.left.group[0], Atom))
        self.assertEqual(ast.left.left.group[0].value, "e")
        self.assertTrue(isinstance(ast.left.left.child, And))
        self.assertTrue(isinstance(ast.left.left.child.left, Atom))
        self.assertEqual(ast.left.left.child.left.value, "p")
        self.assertTrue(isinstance(ast.left.left.child.right, Atom))
        self.assertEqual(ast.left.left.child.right.value, "q")
        self.assertTrue(isinstance(ast.left.right, K))
        self.assertTrue(isinstance(ast.left.right.agent, Atom))
        self.assertEqual(ast.left.right.agent.value, "a")
        self.assertTrue(isinstance(ast.left.right.child, Atom))
        self.assertEqual(ast.left.right.child.value, "p")
        
        self.assertTrue(isinstance(ast.right, CEG))
        self.assertEqual(len(ast.right.group), 1)
        self.assertTrue(isinstance(ast.right.group[0], Atom))
        self.assertEqual(ast.right.group[0].value, "d")
        self.assertTrue(isinstance(ast.right.child, nC))
        self.assertEqual(len(ast.right.child.group), 2)
        self.assertTrue(isinstance(ast.right.child.group[0], Atom))
        self.assertEqual(ast.right.child.group[0].value, "g")
        self.assertTrue(isinstance(ast.right.child.group[1], Atom))
        self.assertEqual(ast.right.child.group[1].value, "e")
        self.assertTrue(isinstance(ast.right.child.child, Implies))
        self.assertTrue(isinstance(ast.right.child.child.left, Atom))
        self.assertEqual(ast.right.child.child.left.value, "c")
        self.assertTrue(isinstance(ast.right.child.child.right, EF))
        self.assertTrue(isinstance(ast.right.child.child.right.child, Atom))
        self.assertEqual(ast.right.child.child.right.child.value, "k")


"""
ATLK with partial observability evaluation functions.
Partial strategies implementation.
"""

from functools import reduce

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp
from pynusmv.exception import PyNuSMVError

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)

from ..atlkFO.eval import (fair_states, ex, eg, eu, nk, ne, nd, nc)

from . import config

import gc

# A dictionary to keep track of number of strategies for improved algorithm
__strategies = {} 
# A dictionary to keep track of number of filterings for improved algorithm
__filterings = {} 
# A dictionary to keep track of number of ignorings for improved algorithm
__ignorings = {}

class StrategyFound(Exception):
    """Exception used to tell that a strategy has been found for all states."""
    def __init__(self, satisfied):
        self.satisfied = satisfied


def evalATLK(fsm, spec, states=None, variant="SF", semantics="group"):
    """
    Return the BDD representing the subset of "states" of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based ATLK specification
    states -- a BDD-based set of states of fsm (if None, the initial states)
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies
                 then filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning
                 states then splitting all remaining actions into uniform
                 strategies, then filtering final winning states.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    If semantics is not in {"group", "individual"}, the "group" semantics is
    used.
    """
    
    if not states:
        states = fsm.init
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager) & states
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager) & states
        
    elif type(spec) is Init:
        return fsm.init & states
        
    elif type(spec) is Reachable:
        return fsm.reachable_states & states
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value) & states
        
    elif type(spec) is Not:
        return states - evalATLK(fsm, spec.child, states, variant=variant)
        
    elif type(spec) is And:
        # TODO We can be smarter by evaluating the second member
        # only on states satisfying the first member
        return (evalATLK(fsm, spec.left, states, variant=variant)
                & evalATLK(fsm, spec.right, states, variant=variant))
        
    elif type(spec) is Or:
        # TODO We can be smarter by evaluating the second member
        # only on states not satisfying the first member
        return (evalATLK(fsm, spec.left, states, variant=variant)
                | evalATLK(fsm, spec.right, states, variant=variant))
        
    elif type(spec) is Implies:
        # TODO We can be smarter by evaluating the second member
        # only on states satisfying the first member
        # p -> q = ~p | q
        p = spec.left
        q = spec.right
        newspec = Or(Not(p), q)
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is Iff:
        # p <-> q = (p & q) | (~p & ~q)
        p = spec.left
        q = spec.right
        newspec = Or(And(p, q), And(Not(p), Not(q)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is EX:
        phi = evalATLK(fsm, spec.child, fsm.post(states), variant=variant)
        return ex(fsm, phi) & states
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        newspec = Not(EX(Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is EG:
        phi = evalATLK(fsm, spec.child, reach(fsm, states), variant=variant)
        return eg(fsm, phi) & states
        
    elif type(spec) is AG:
        # AG p = ~EF ~p (= ~E[ true U ~p ])
        newspec = Not(EF(Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is EU:
        phi1 = evalATLK(fsm, spec.left, reach(fsm, states), variant=variant)
        phi2 = evalATLK(fsm, spec.right, reach(fsm, states), variant=variant)
        return eu(fsm, phi1, phi2) & states
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = spec.left
        q = spec.right
        newspec = Not(Or(EU(Not(q), And(Not(p), Not(q))), EG(Not(q))))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is EF:
        # EF p = E[ true U p ]
        newspec = EU(TrueExp(), spec.child)
        return evalATLK(fsm, newspec, states, variant=variant)  
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        newspec = Not(EG(Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        p = spec.left
        q = spec.right
        newspec = Or(EU(p, q), EG(p))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = spec.left
        q = spec.right
        newspec = Not(EU(Not(q), And(Not(p), Not(q))))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is nK:
        agent = frozenset({spec.agent.value})
        phi = evalATLK(fsm, spec.child,
                       fsm.equivalent_states(states, agent),
                       variant=variant)
        return states & nk(fsm, spec.agent.value, phi)
        
    elif type(spec) is K:
        # K<'a'> p = ~nK<'a'> ~p
        newspec = Not(nK(spec.agent, Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is nE:
        agents = frozenset([a.value for a in spec.group])
        phi = Eequiv(fsm, states, agents)
        return states & ne(fsm, agents, phi)
        
    elif type(spec) is E:
        # E<g> p = ~nE<g> ~p
        newspec = Not(nE(spec.group, Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is nD:
        agents = frozenset([a.value for a in spec.group])
        phi = Dequiv(fsm, states, agents)
        return states & nd(fsm, agents, phi)
        
    elif type(spec) is D:
        # D<g> p = ~nD<g> ~p
        newspec = Not(nD(spec.group, Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is nC:
        agents = frozenset([a.value for a in spec.group])
        phi = Cequiv(fsm, states, agents)
        return states & nc(fsm, agents, phi)
        
    elif type(spec) is C:
        # C<g> p = ~nC<g> ~p
        newspec = Not(nC(spec.group, Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = Not(CEX(spec.group, Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = Not(CEF(spec.group, Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = Not(CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right))))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = Not(CEG(spec.group, Not(spec.child)))
        return evalATLK(fsm, newspec, states, variant=variant)
        
    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = Not(CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right))))
        return evalATLK(fsm, newspec, states, variant=variant)
                   
    elif type(spec) in {CEX, CEF, CEG, CEW, CEU}:
        if variant == "SF":
            return eval_strat(fsm, spec, states, semantics=semantics)
        elif variant == "FS":
            return eval_strat_improved(fsm, spec, states, semantics=semantics)
        elif variant == "FSF":
            print("[WARNING] evalATLK: no FSF evaluation for now",
                  " basic evaluation used instead.")
            #return eval_strat_FSF(fsm, spec, states)
            return eval_strat(fsm, spec, states, semantics=semantics)
        else:
            return eval_strat(fsm, spec, states, semantics=semantics)
        
    else:
        # TODO Generate error
        print("[ERROR] evalATLK: unrecognized specification type", spec)
        return None


def reach(fsm, states):
    """
    Return the set of states reachable from states in fsm.
    
    fsm -- a MAS representing the system
    states -- a BDD representing a set of states of fsm.
    """
    return fp(lambda Z: states | fsm.post(Z), BDD.false(fsm.bddEnc.DDmanager))
    

def Eequiv(fsm, states, agents):
    """
    Return the set of fsm that equivalent to states w.r.t. group knowledge of
    agents.
    
    fsm -- a MAS representing the system
    states -- a BDD representing a set of states of fsm
    agents -- a set of agents names of fsm.
    """
    result = BDD.false(fsm.bddEnc.DDmanager)
    for agent in group:
        result = result | fsm.equivalent_states(states, frozenset({agent}))
    return result
    

def Dequiv(fsm, states, agents):
    """
    Return the set of fsm that equivalent to states w.r.t. distributed
    knowledge of agents.
    
    fsm -- a MAS representing the system
    states -- a BDD representing a set of states of fsm
    agents -- a set of agents names of fsm.
    """
    return fsm.equivalent_states(states, frozenset({agents}))
    

def Cequiv(fsm, states, agents):
    """
    Return the set of fsm that equivalent to states w.r.t. group knowledge of
    agents.
    
    fsm -- a MAS representing the system
    states -- a BDD representing a set of states of fsm
    agents -- a set of agents names of fsm.
    """
    return fp(lambda Z: Eequiv(fsm, states | Z, agents),
              BDD.false(fsm.bddEnc.DDmanager))

def fair_states_sub(fsm, subsystem=None):
    """
    Return the set of fair states of the subsystem.
    
    fsm -- the model;
    subsystem -- if not None, the subsystem defined as a set of state/inputs
                 pairs.
    """
    
    if len(fsm.fairness_constraints) == 0:
        return BDD.true(fsm.bddEnc.DDmanager)
    else:
        return eg_sub(fsm, BDD.true(fsm.bddEnc.DDmanager), subsystem=subsystem)

def reachable_sub(fsm, init=None, subsystem=None):
    """
    Return the set of states reachable from init in the subsystem.
    
    fsm -- the model;
    init -- if not None, the initial states
            otherwise, the initial states of the model;
    subsystem -- if not None, the subsystem to consider
                 otherwise, the original system.
    """
    
    if init is None:
        init = fsm.init
    if subsystem is None:
        subsystem = BDD.true(fsm.bddEnc.DDmanager)
    
    old = BDD.false(fsm.bddEnc.DDmanager)
    new = init & fsm.bddEnc.statesMask
    while old != new:
        old = new
        new = ((old | fsm.post(old, subsystem=subsystem)) &
               fsm.bddEnc.statesMask)
    return new
    
def ex_sub(fsm, phi, subsystem=None):
    """
    Return the set of states of fsm satisfying EX phi in subsystem.
    
    fsm -- a MAS representing the system
    phi -- a BDD representing the set of states of fsm satisfying phi
    subsystem -- if not None, the subsystem defined as a set of state/inputs
                 pairs.
    """
    
    if subsystem is None:
        subsystem = BDD.true(fsm.bddEnc.DDmanager)
    phi = phi.forsome(fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask
    
    return fsm.pre(phi & fair_states_sub(fsm, subsystem=subsystem),
                   subsystem=subsystem)
    
    
def eg_sub(fsm, phi, subsystem=None):
    """
    Return the set of states of fsm satisfying EG phi in subsystem.
    
    fsm -- a MAS representing the system
    phi -- a BDD representing the set of states of fsm satisfying phi
    subsystem -- if not None, the subsystem defined as a set of state/inputs
                 pairs.
    """
    
    if subsystem is None:
        subsystem = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi.forsome(fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask
    
    if len(fsm.fairness_constraints) == 0:
        return (fp(lambda Z : phi & fsm.pre(Z, subsystem=subsystem),
                   BDD.true(fsm.bddEnc.DDmanager)).
                   forsome(fsm.bddEnc.inputsCube))
    else:
        def inner(Z):
            res = phi
            for f in fsm.fairness_constraints:
                res = res & fsm.pre(fp(lambda Y : (Z & f) |
                                       (phi & fsm.pre(Y, subsystem=subsystem)),
                                       BDD.false(fsm.bddEnc.DDmanager)),
                                    subsystem=subsystem)
            return res
        return (fp(inner, BDD.true(fsm.bddEnc.DDmanager))
                .forsome(fsm.bddEnc.inputsCube))
    
    
def eu_sub(fsm, phi, psi, subsystem=None):
    """
    Return the set of states of fsm satisfying E[ phi U psi ] in subsystem.
    
    fsm -- a MAS representing the system
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    subsystem -- if not None, the subsystem defined as a set of state/inputs
                 pairs.
    """
    
    phi = phi.forsome(fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask
    psi = psi.forsome(fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask
    return fp(lambda X : (psi & fair_states_sub(fsm, subsystem=subsystem) &
                          fsm.reachable_states) |
                               (phi & ex_sub(fsm, X, subsystem=subsystem)),
                    BDD.false(fsm.bddEnc.DDmanager))


def cex_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> X phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask
    
    return fsm.pre_strat_si(phi | nfair_gamma_si(fsm, agents, strat),
                            agents, strat)
    

def ceu_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying
    <agents>[phi U psi] under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_strat_si(Z, agents, strat)),
                  BDD.false(fsm.bddEnc.DDmanager))
    else:
        nfair = nfair_gamma_si(fsm, agents, strat)
        def inner(Z):
            res = psi
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (phi | psi | nfair) &
                                                 (Z | nf) &
                                                 (psi |
                                                  fsm.pre_strat_si(Y, agents,
                                                                   strat)
                                                 ),
                                              BDD.true(fsm.bddEnc.DDmanager)),
                                              agents, strat)
            return (psi | phi | nfair) & res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    

def cew_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying
    <agents>[phi W psi] under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (psi | phi | nfair) &
                         (psi | fsm.pre_strat_si(Y, agents, strat)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
def ceg_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> G phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (phi | nfair) & fsm.pre_strat_si(Y, agents, strat),
              BDD.true(fsm.bddEnc.DDmanager))


def nfair_gamma_si(fsm, agents, strat=None):
    """
    Return the set of state/inputs pairs of strat
    in which agents can avoid a fair path in strat.
    If strat is None, it is considered true.
    
    fsm -- the model
    agents -- a list of agents names
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    if len(fsm.fairness_constraints) == 0:
        return BDD.false(fsm.bddEnc.DDmanager)
    else:
        def inner(Z):
            res = BDD.false(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (Z | nf) &
                                                 fsm.pre_strat_si(Y, agents,
                                                                  strat),
                                             BDD.true(fsm.bddEnc.DDmanager)),
                                             agents, strat)
            return res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
        
        
def split_reach(fsm, agents, pustrat, subsystem=None, semantics="group"):
    """
    Return the set of maximal uniform strategies extending pustrat.
    
    fsm -- a MAS representing the system;
    agents -- a set of agents names;
    pustrat -- a partial uniform strategy represented as BDD-based set of
               moves (state/action pairs);
    subsystem -- the subsystem in which building the strategies;
                 the full system if None.
    semantics -- the semantic to use (group or individual);
                 if not "individual", "group" semantics is used
    
    """
    
    if subsystem is None:
        subsystem = BDD.true(fsm.bddEnc.DDmanager)
    
    # Get new states
    new = (fsm.post(pustrat, subsystem) -
           pustrat.forsome(fsm.bddEnc.inputsCube)).forsome(
                                 fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask

    if new.is_false():
        yield pustrat
    
    else:
        for npustrat in split(fsm,
                              new & fsm.protocol(agents),
                              agents,
                              pustrat & subsystem,
                              semantics=semantics):
            for strat in split_reach(fsm, agents, pustrat | npustrat,
                                     subsystem, semantics=semantics):
                yield strat
        

def get_equiv_class(fsm, gamma, state):
    """
    Return the equivalence class for gamma state belongs to. The equivalence
    class depends on semantics; The distributed knowledge based
    equivalence fo gamma containing state is returned.
    
    fsm -- the model
    gamma -- a set of agents of fsm
    """
    return fsm.equivalent_states(state, gamma) & fsm.reachable_states


def is_conflicting(fsm, eqclass, gamma):
    """
    Return whether the given equivalence class for gamma is conflicting or not.
    
    fsm -- the model
    eqclass -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    
    eqclass must be a distributed knowledge based equivalence class.
    
    """
    si = fsm.pick_one_state_inputs(eqclass)
    ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
    return (eqclass - (eqclass & si.forsome(fsm.bddEnc.statesCube |
                                            ngamma_cube))).isnot_false()


def split_conflicting(fsm, eqclass, gamma, pustrat):
    """
    Split the given equivalence class for gamma into non conflicting
    sub-strategies.
    Restrict to actions allowed in pustrat, that is, if a state of strats
    is equivalent to one of pustrat, only allows actions given by pustrat.
    
    fsm -- the model
    eqclass -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    pustrat -- a partial uniform strategy represented
               by a BDD-based set of moves
    
    eqclass must be a distributed knowledge based equivalence class.
    
    Generate all splitted non conflicting sub-strategies.
    
    """
    ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
    # Split eqclass into non-conflicting subsets
    while eqclass.isnot_false():
        si = fsm.pick_one_state_inputs(eqclass)
        ncss = eqclass & si.forsome(fsm.bddEnc.statesCube | ngamma_cube)
        eqclass = eqclass - ncss
        
        # Take pustrat into account
        # eq is the set of states equivalent to ncss states
        eq = get_equiv_class(fsm, gamma,
                             ncss.forsome(fsm.bddEnc.inputsCube))
        if (eq & pustrat).isnot_false():
            # Some states are equivalent in ncss and pustrat
            
            # Check that the strategy is compatible
            lstrat = eq & pustrat # the strategy in pustrat
            action = lstrat.forsome(fsm.bddEnc.statesCube | ngamma_cube)
            # The strategy is compatible if ncss proposes the same
            # action as lstrat
            if (action & ncss).isnot_false():
                yield ncss
            # Otherwise the strategy is incompatible, discard it
        
        # No new state is equivalent to a pustrat one
        else:
            yield ncss


def split_one(fsm, strats, gamma, pustrat):
    """
    Split one equivalence class of strats and return triples composed of
    the common non-conflicting part already encountered,
    a split and the rest of strats to split.
    Restrict to actions allowed in pustrat, that is, if a state of strats
    is equivalent to one of pustrat, only allows actions given by pustrat.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    pustrat -- a partial uniform strategy represented
               by a BDD-based set of moves
    
    Return a generator of all triples of common parts, splits
    and rest of strats.
    
    """
    if strats.is_false():
        yield (strats, strats, strats)
    else:
        common = BDD.false(fsm.bddEnc.DDmanager)
        
        while strats.isnot_false():
            # Get one equivalence class
            si = fsm.pick_one_state_inputs(strats)
            s = si.forsome(fsm.bddEnc.inputsCube)
            eqs = get_equiv_class(fsm, gamma, s)
            eqcl = strats & eqs
            
            # Remove it from strats
            strats = strats - eqcl
            
            # The current equivalence class is conflicting
            if is_conflicting(fsm, eqcl, gamma):
                for non_conflicting in split_conflicting(fsm, eqcl, gamma,
                                                         pustrat):
                    yield (common, non_conflicting, strats)
                return
            
            else:
                # Add equivalence class to common
                common = common | eqcl
        
        # strats is false, everything is in common
        yield (common, strats, strats)


def split(fsm, strats, gamma, pustrat=None, semantics="group"):
    """
    Split strats into all its non-conflicting greatest subsets.
    Restrict to actions allowed in pustrat, that is, if a state of strats
    is equivalent to one of pustrat, only allows actions given by pustrat.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    pustrat -- a partial uniform strategy represented
               by a BDD-based set of moves
    semantics -- the semantic to use for splitting (group or individual)
    
    Return a generator of all non-conflicting greatest subsets of strats.
    
    """
    if semantics == "individual":
        if len(gamma) <= 0:
            yield strats
        else:
            agent = gamma[0]
            gamma = gamma[1:]
            for strat in split(fsm, strats, [agent], pustrat=pustrat,
                               semantics="group"):
                for sstrat in split(fsm, strat, gamma, pustrat=pustrat,
                                    semantics="individual"):
                    yield sstrat
    else:
        if not pustrat:
            pustrat = BDD.false(fsm.bddEnc.DDmanager)
        
        strats = strats & fsm.bddEnc.statesInputsMask
        pustrat = pustrat & fsm.bddEnc.statesInputsMask
        
        if strats.is_false():
            yield strats
        else:
            # Split one equivalence class
            for common, splitted, rest in split_one(fsm, strats, gamma,
                                                    pustrat):
                for strat in split(fsm, rest, gamma, pustrat=pustrat):
                    yield (common | strat | splitted)


def all_equiv_sat(fsm, winning, agents, semantics="group"):
    """
    Return the states s of winning such that all states indistinguishable
    from s by agents are in winning.
    
    fsm -- a MAS representing the system;
    winning -- a BDD representing a set of states of fsm;
    agents -- a set of agents;
    semantics -- the semantics to use (group or individual); if not
                 "individual", "group" semantics is used.
    
    """
    if semantics == "individual":
        equiv_sat = BDD.true(fsm.bddEnc.DDmanager)
        for agent in agents:
            equiv_sat &= all_equiv_sat(fsm, winning, {agent},
                                       semantics="group")
        return equiv_sat
    else:
        # Get states for which all states belong to winning
        # wineq is the set of states for which all equiv states are in winning
        nwinning = ~winning & fsm.bddEnc.statesInputsMask
        return ~(fsm.equivalent_states(nwinning &
                 fsm.reachable_states, frozenset(agents))) & winning
             
             
def filter_strat(fsm, spec, states, strat=None, variant="SF",
                 semantics="group"):
    """
    Returns the subset SA of strat (or the whole system if strat is None),
    state/action pairs of fsm, such that there is a strategy to satisfy spec
    in fsm.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator;
            the operator is CEX, CEG, CEF, CEU or CEW.
    states -- the set of states for which the filtering matters;
    strat -- the subset of the system to consider.
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies
                 then filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning
                 states then splitting all remaining actions into uniform
                 strategies, then filtering final winning states.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    if strat is None:
        strat = BDD.true(fsm.bddEnc.DDmanager)
        # We can be more restrictive here: strat = fsm.protocol(agents)
        # But it is not a problem since the strat is restricted later to
        # protocol
    
    # Filtering
    if type(spec) is CEX:
        phi = evalATLK(fsm, spec.child, fsm.post(states & strat),
                       variant=variant, semantics=semantics)
        winning = cex_si(fsm, agents, phi, strat)

    elif type(spec) is CEG:
        phi = evalATLK(fsm, spec.child, strat.forsome(fsm.bddEnc.inputsCube), 
                       variant=variant, semantics=semantics)
        winning = ceg_si(fsm, agents, phi, strat)

    elif type(spec) is CEU:
        phi1 = evalATLK(fsm, spec.left, strat.forsome(fsm.bddEnc.inputsCube), 
                       variant=variant, semantics=semantics)
        phi2 = evalATLK(fsm, spec.right, strat.forsome(fsm.bddEnc.inputsCube), 
                       variant=variant, semantics=semantics)
        winning = ceu_si(fsm, agents, phi1, phi2, strat)

    elif type(spec) is CEF:
        # <g> F p = <g>[true U p]
        phi = evalATLK(fsm, spec.child, strat.forsome(fsm.bddEnc.inputsCube), 
                       variant=variant, semantics=semantics)
        winning = ceu_si(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
                         phi, strat)

    elif type(spec) is CEW:
        phi1 = evalATLK(fsm, spec.left, strat.forsome(fsm.bddEnc.inputsCube), 
                       variant=variant, semantics=semantics)
        phi2 = evalATLK(fsm, spec.right, strat.forsome(fsm.bddEnc.inputsCube), 
                       variant=variant, semantics=semantics)
        winning = cew_si(fsm, agents, phi1, phi2, strat)
    
    return winning & fsm.bddEnc.statesInputsMask & fsm.protocol(agents)


def agents_in_group(fsm, group):
    """
    Return the list of agents in the given group in fsm. If group is not a
    group of fsm, group is returned alone.
    
    This procedure recursively searches for all basic agents in the groups
    possibly composing group.
    
    fsm -- the model;
    group -- a group or agent name of the model.
    """
    
    if group in fsm.groups:
        agents = []
        for agent in fsm.groups[group]:
            if agent in fsm.groups:
                agents += [a for a in agents_in_group(fsm, agent)
                           if a not in agents]
            else:
                agents.append(agent)
    else:
        agents = [group]
    return agents


def agents_in_list(fsm, agents):
    """
    Return the list of agents in the given list of groups and agents.
    
    fsm -- the model;
    agents -- an iterable of groups and agents.
    """
    result = []
    for agent in agents:
        result += [a for a in agents_in_group(fsm, agent)
                   if a not in result]
    return result


def partial_strategies(fsm, states, agents, subsystem=None, semantics="group"):
    """
    Generate all partial strategies adequate for states in subsystem.
    
    fsm -- a MAS representing the system;
    states -- a BDD representing a set of states of fsm;
    agents -- a set of agents of fsm;
    subsystem -- if not None, a set of moves of agents;
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
    """
    for pustrat in split(fsm, states & subsystem, agents, semantics=semantics):
        for strat in split_reach(fsm,
                                 agents,
                                 pustrat,
                                 subsystem,
                                 semantics=semantics):
            yield strat


def eval_strat(fsm, spec, states, semantics="group"):
    """
    Return the BDD representing the subset of states satisfying spec.
    spec is a strategic operator <G> pi.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic (<g>)
            operator;
    states -- a BDD representing a set of states of fsm.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
    
    """
    
    agents = [atom.value for atom in spec.group]
    
    if semantics == "individual":
        agents = agents_in_list(fsm, agents)
    
    if config.debug:
        print("Evaluating partial strategies for ", spec)
    
    # Extend with equivalent states
    if semantics == "individual":
        for agent in agents:
            states = states | get_equiv_class(fsm, {agent}, states)
    else:
        states = get_equiv_class(fsm, agents, states)
    
    # Pre-filtering out losing states and actions
    if config.partial.filtering:
        # FIXME This triggers full partial strategies exploration when
        # a sub-formula is a strategic one. Should perform only full obs
        # approximation for strategic sub-formulas as well.
        # Can be done with semantics="full obs"
        subsystem = filter_strat(fsm, spec, states, variant="SF",
                                 semantics=semantics)
    else:
        subsystem = fsm.protocol(agents)
    # if filtering is enabled, subsystem is the part of the system in which
    # states can win
    
    # if filtering is enabled, we can remove, from states, the set of states
    # for which not all equivalent states are in subsystem. Indeed, if a state
    # is not in subsystem, there is not strategy in this state to win,
    # thus no uniform strategy to win, thus its entire equivalence class
    # cannot win
    states = (states & all_equiv_sat(fsm,
                                     subsystem.forsome(fsm.bddEnc.inputsCube), 
                                     agents, semantics=semantics))
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    
    all_states = states
    
    nbstrats = 0
    
    while all_states.isnot_false():
        
        # ----- Separation of state space -------------------------------------
        
        # Put a subset of all_states in states and remove it from all_states
        if config.partial.separation.type is None:
            states = all_states
            
        if config.partial.separation.type == "random":
            state = fsm.pick_one_state(all_states)
            states = (get_equiv_class(fsm, agents, state, semantics=semantics)
                      & all_states)
                      
        if config.partial.separation.type == "reach":
            reached = fsm.init
            while (reached & all_states).is_false():
                # Should not loop infinitely since we only are interested in
                # reachable states
                # WARNING We cannot restrict to subsystem to get the post-image
                # since we are not sure that all_states are reachable in the
                # subsystem anymore
                reached = reached | fsm.post(reached)
            state = fsm.pick_one_state(reached & all_states)
            states = (get_equiv_class(fsm, agents, state, semantics=semantics)
                      & all_states)
        
        # ---------------------------------------------------------------------
        
        
        # Remove states from all_states
        all_states = all_states - states
    
        orig_states = states
        remaining = states
    
        while remaining.isnot_false():
            # Extend states with equivalent ones
            if semantics == "individual":
                states = remaining
                for agent in agents:
                    states = states | get_equiv_class(fsm, {agent}, states)
            else:
                states = get_equiv_class(fsm, agents, remaining)
        
            remaining_size = fsm.count_states(remaining)
            
            # Go through all strategies
            for strat in partial_strategies(fsm,
                                            states,
                                            agents,
                                            subsystem=subsystem,
                                            semantics=semantics):
                # Check the strategy
                nbstrats += 1
                winning = (filter_strat(fsm, spec, states, strat,
                                        variant="SF", semantics=semantics).
                           forsome(fsm.bddEnc.inputsCube))
                old_sat = sat
                sat = sat | (all_equiv_sat(fsm, winning, agents,
                                           semantics=semantics) & orig_states)
            
            
                # ----- EARLY TERMINATION -------------------------------------
            
                # Early termination if sat contains all requested states
                if config.partial.early.type == "full" and orig_states <= sat:
                    remaining = BDD.false(fsm.bddEnc.DDmanager)
                    break
            
                # Early termination of sat grows
                if config.partial.early.type == "partial" and old_sat < sat:
                    if config.debug:
                        print("Partial strategies: sat grows ({} strateg{})"
                              .format(nbstrats,
                                      "ies" if nbstrats > 1 else "y"))
                          
                    remaining = remaining - sat
                    break
            
                # Early termination if remaining states decrease enough
                if config.partial.early.type == "threshold":
                    remaining = remaining - sat
                    rem_count = fsm.count_states(remaining)
                    if (rem_count <= 
                        remaining_size * config.partial.early.threshold):
                        if config.debug:
                            print("Partial strategies:",
                                  "remaining states decrease:",
                                  "{} => {}".format(remaining_size, rem_count))
                              
                        remaining_size = rem_count
                        break
            
                # -------------------------------------------------------------
                
                
                if config.debug and nbstrats % 1000 == 0:
                    print("Partial strategies: {} strateg{} checked so far"
                          .format(nbstrats, "ies" if nbstrats > 1 else "y"))
                
                
                # ----- Garbage collection ------------------------------------
                if (config.garbage.type == "each" or
                    (config.garbage.type == "step"
                        and nbstrats % config.garbage.step == 0)):
                    gc.collect()
        
            else:
                # All strategies have been checked, the remaining states do not
                # satisfy the specification
                remaining = BDD.false(fsm.bddEnc.DDmanager)
        
        
        # Remove from remaining subsets to check the ones we already have the
        # truth value
        all_states = all_states - sat
    
    # DEBUG Print number of strategies
    if config.debug:
        print("Partial strategies: {} strateg{} generated"
              .format(nbstrats, "ies" if nbstrats > 1 else "y"))
    
    return sat


def eval_univ(fsm, spec, states, subsystem=None, variant="FS",
              semantics="group"):
    """
    Given spec a universal CTL formula, return the subset of states satisfying
    spec in the given subsystem.
    
    fsm -- the model;
    spec -- a universal CTL formula;
    states -- a subset of the states of the model;
    subsystem -- if not None, the subsystem in which evaluate the formula;
    variant -- the variant to evaluate sub-formulas.
    semantics -- the semantic to use for splitting (group or individual);
                 if not "individual", "group" semantics is used
    """
    
    if subsystem is None:
        subsystem = BDD.true(fsm.bddEnc.DDmanager)
    
    interesting = subsystem.forsome(fsm.bddEnc.inputsCube)
    
    if type(spec) is AX:
        phi = evalATLK(fsm, spec.child,
                       fsm.post(states, subsystem=subsystem),
                       variant=variant, semantics=semantics)
        winning = ~ex_sub(fsm, ~phi, subsystem)
    
    elif type(spec) is AG:
        phi = evalATLK(fsm, spec.child,
                       interesting,
                       variant=variant, semantics=semantics)
        winning = ~eu_sub(fsm, BDD.true(fsm.bddEnc.DDmanager), ~phi, subsystem)
    
    elif type(spec) is AU:
        phi1 = evalATLK(fsm, spec.left,
                        interesting,
                        variant=variant, semantics=semantics)
        phi2 = evalATLK(fsm, spec.right,
                        interesting,
                        variant=variant, semantics=semantics)
        winning = ~(eu_sub(fsm, ~phi2, ~phi1 & ~phi2, subsystem) |
                    eg_sub(fsm, ~phi2, subsystem))
    
    elif type(spec) is AF:
        phi = evalATLK(fsm, spec.child,
                       interesting,
                       variant=variant, semantics=semantics)
        winning = ~eg_sub(fsm, ~phi, subsystem)
    
    elif type(spec) is AW:
        phi1 = evalATLK(fsm, spec.left,
                        interesting,
                       variant=variant, semantics=semantics)
        phi2 = evalATLK(fsm, spec.right,
                        interesting,
                        variant=variant, semantics=semantics)
        winning = ~eu_sub(fsm, ~phi2, ~phi1 & ~phi2, subsystem)
    
    return winning & states & fsm.bddEnc.statesMask
    

def strat_to_univ(spec):
    """
    Return the universal formula sharing the same path operator
    as the strategic formula spec.
    
    spec -- a formula with a strategic top operator
    """
    if type(spec) is CEX:
        return AX(spec.child)
    elif type(spec) is CEF:
        return AF(spec.child)
    elif type(spec) is CEG:
        return AG(spec.child)
    elif type(spec) is CEU:
        return AU(spec.left, spec.right)
    elif type(spec) is CEW:
        return AW(spec.left, spec.right)

def eval_strat_improved(fsm, spec, states, semantics="group"):
    """
    Return the BDD representing the subset of states satisfying spec.
    spec is a strategic operator <G> pi.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic (<g>)
            operator;
    states -- a BDD representing a set of states of fsm.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
    """
    
    assert(len(config.partial.alternate.type & {"univ", "strat"}) > 0)
    
    if config.debug:
        print("Evaluating partial strategies for ", spec, "(early)")
    
    __strategies[spec] = 0
    __filterings[spec] = 0
    __ignorings[spec] = 0
    
    gamma = [atom.value for atom in spec.group]
    
    if semantics == "individual":
        gamma = agents_in_list(fsm, gamma)
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    
    if semantics == "individual":
        for agent in gamma:
            states = states | get_equiv_class(fsm, {agent}, states)
    else:
        states = get_equiv_class(fsm, gamma, states)
    
    # Pre-filtering out losing states and actions
    if config.partial.filtering:
        # FIXME This triggers full partial strategies exploration when
        # a sub-formula is a strategic one. Should perform only full obs
        # approximation for strategic sub-formulas as well.
        # Can be done with semantics="full obs"
        subsystem = filter_strat(fsm, spec, states, variant="FS",
                                 semantics=semantics)
    else:
        subsystem = fsm.protocol(gamma)
    # if filtering is enabled, subsystem is the part of the system in which
    # states can win
    
    # if filtering is enabled, we can remove, from states, the set of states
    # for which not all equivalent states are in subsystem. Indeed, if a state
    # is not in subsystem, there is not strategy in this state to win,
    # thus no uniform strategy to win, thus its entire equivalence class
    # cannot win
    states = (states & all_equiv_sat(fsm,
                                     subsystem.forsome(fsm.bddEnc.inputsCube), 
                                     gamma, semantics=semantics))
    
    remaining = states
    
    for pstrat in split(fsm, states & fsm.protocol(gamma), gamma,
                        pustrat=subsystem,
                        semantics=semantics):
        newsat = eval_strat_alternate(fsm, spec, gamma, remaining, pstrat,
                                      subsystem=subsystem,
                                      semantics=semantics)
        sat = sat | newsat
        remaining = remaining - sat
        if sat == states:
            break # Early termination
    
    if config.debug:
        # The number of maximal partial strategies computed
        nbs = __strategies[spec]
        # The number of strategic and universal evaluations of
        # (not necessarily maximal) partial strategies computed
        nbf = __filterings[spec]
        # The number of non-maximal partial strategies ignored
        # because a decision could be made before reaching a maximal strategy
        nbi = __ignorings[spec]
        print("Early: {} strateg{}, {} filtering{}, {} ignoring{} computed."
              .format(nbs, "y" if nbs <= 1 else "ies",
                      nbf, "" if nbf <= 1 else "s",
                      nbi, "" if nbi <= 1 else "s"))
    
    return sat

def eval_strat_alternate(fsm, spec, gamma, states, pstrat, subsystem,
                         semantics="group"):
    """
    Return the BDD representing the subset of states s' such that there exists
    a uniform maximal partial strategy extending pstrat that is winning for
    spec = <gamma> psi in [s']_gamma.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic (<g>)
            operator;
    gamma -- the list of agents of spec (expanded to single agents in the case
             of individual semantics); must correspond to the agents of spec;
    states -- a BDD representing a set of states of fsm;
    pstrat -- a uniform partial strategy reachable from states.
    subsystem -- the subsystem to search for strategies in.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
    """
    
    fullpstrat = (pstrat |
                  (fsm.protocol(gamma) -
                   pstrat.forsome(fsm.bddEnc.inputsCube)))
    
    # Get losing states
    if "strat" in config.partial.alternate.type:
        fullpstrat = fullpstrat & reachable_sub(fsm, states, fullpstrat)
        
        notlosing = filter_strat(fsm, spec, states, fullpstrat, variant="FS",
                                 semantics=semantics)
        notlosing = notlosing.forsome(fsm.bddEnc.inputsCube)
        notlosing = notlosing & states
        lose = states - all_equiv_sat(fsm, notlosing, gamma,
                                      semantics=semantics)
        
        states = states - lose
    else:
        lose = BDD.false(fsm.bddEnc.DDmanager)
    
    # Get winning states
    if "univ" in config.partial.alternate.type:
        fullpstrat = fullpstrat & reachable_sub(fsm, states, fullpstrat)
        
        swin = eval_univ(fsm, strat_to_univ(spec), states, fullpstrat,
                           variant="FS", semantics=semantics)
        win = all_equiv_sat(fsm, swin, gamma, semantics=semantics) & states
        
        states = states - win
    else:
        win = BDD.false(fsm.bddEnc.DDmanager)
    
    __filterings[spec] += 1
    
    if config.debug and __filterings[spec] % 1000 == 0:
        print("Partial strategies (FS): {} filtering{} done so far"
              .format(__filterings[spec],
                      "s" if __filterings[spec] > 1 else ""))
    
    # Check whether there are still states to decide
    if states.is_false():
        __ignorings[spec] += 1
        
        if config.debug and __ignorings[spec] % 1000 == 0:
            print("Partial strategies (FS): {} ignoring{} done so far"
                  .format(__ignorings[spec],
                          "s" if __ignorings[spec] > 1 else ""))
        
        # Collect to avoid memory overflow
        if (config.garbage.type == "each" or config.garbage.type == "step"
            and __strategies[spec] % config.garbage.step == 0):
            
            gc.collect()
        
        return win
    
    # Expand the partial strategy to decide for the remaining states
    else:
        # Expand the partial strategy: get new reachable states
        new = (fsm.post(pstrat) - 
               pstrat.forsome(fsm.bddEnc.inputsCube)).forsome(
               fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask
        if new.is_false():
            __strategies[spec] += 1

            if config.debug and __strategies[spec] % 1000 == 0:
                print("Partial strategies (FS): {} strateg{} checked so far"
                      .format(__strategies[spec],
                              "ies" if __strategies[spec] > 1 else "y"))
            
            # No new states, pstrat is already maximal
            # Winning states in win if win has some meaning
            # otherwise lose has some meaning and winning states are the others
            if "univ" in config.partial.alternate.type:
                return win
            else:
                return states - lose
        
        remaining = states
        
        # Split the new moves
        for npstrat in split(fsm, new & fsm.protocol(gamma), gamma,
                             pustrat=pstrat & subsystem,
                             semantics=semantics):
            newwin = eval_strat_alternate(fsm, spec, gamma, remaining,
                                          pstrat | npstrat,
                                          subsystem=subsystem,
                                          semantics=semantics)
            win = win | newwin
            remaining = remaining - win
            if win == states:
                # Early termination when we know the truth value of all states
                break
        return win

# -----------------------------------------------------------------------------
# CACHING FUNCTIONS AND MANIPULATIONS
# -----------------------------------------------------------------------------

__evalATLK_cache = {}
__orig_evalATLK = evalATLK
def __cached_evalATLK(fsm, spec, states=None, variant="SF", semantics="group"):
    if config.partial.caching:
        if states is None:
            states = fsm.init
        
        if (fsm, spec, semantics) in __evalATLK_cache.keys():
            sat, unsat = __evalATLK_cache[(fsm, spec, semantics)]
        else:
            false = BDD.false(fsm.bddEnc.DDmanager)
            sat, unsat = false, false

        remaining = states - (sat + unsat)
    
        if remaining.isnot_false():
            remsat = __orig_evalATLK(fsm, spec, remaining, variant,
                                     semantics=semantics)
            remunsat = remaining - remsat
            __evalATLK_cache[(fsm, spec, semantics)] = (sat + remsat,
                                                        unsat + remunsat)
        else:
            remsat = BDD.false(fsm.bddEnc.DDmanager)
    
        return sat + remsat
    
    else:
        return __orig_evalATLK(fsm, spec, states, variant, semantics=semantics)

evalATLK = __cached_evalATLK


"""
ARCTL evaluation functions.
"""

from pyparsing import ParseException

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression

from .parsing import parseArctl

from .ast import (TrueExp, FalseExp,
                  Atom, Not, And, Or, Implies, Iff, 
                  AaF, AaG, AaX, AaU, AaW,
                  EaF, EaG, EaX, EaU, EaW)
                  
                  
def evalArctl_from_string(fsm, spec):
    """
    Parse spec and return a BDD representing the set of states of fsm
    satisfying spec.
    
    Throws a ParseException if spec does not represent an ARCTL formula.
    """
    specs = parseArctl(spec)
    if len(specs) != 1:
        raise ParseException("Multiple parsing results")
    return evalArctl(fsm, specs[0])
    

def evalArctl(fsm, spec):
    """Return a BDD representing the set of states of fsm satisfying spec."""
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalArctl(fsm, spec.child)
        
    elif type(spec) is And:
        return evalArctl(fsm, spec.left) & evalArctl(fsm, spec.right)
        
    elif type(spec) is Or:
        return evalArctl(fsm, spec.left) | evalArctl(fsm, spec.right)
        
    elif type(spec) is Implies:
        return (~evalArctl(fsm, spec.left)) | evalArctl(fsm, spec.right)
        
    elif type(spec) is Iff:
        l = evalArctl(fsm, spec.left)
        r = evalArctl(fsm, spec.right)
        return (l & r) | ((~l) & (~r))
        
    elif type(spec) is AaF:
        return aaf(fsm, evalArctl(fsm, spec.action), evalArctl(fsm, spec.child))
        
    elif type(spec) is AaG:
        return aag(fsm, evalArctl(fsm, spec.action), evalArctl(fsm, spec.child))
        
    elif type(spec) is AaX:
        return aax(fsm, evalArctl(fsm, spec.action), evalArctl(fsm, spec.child))
        
    elif type(spec) is AaU:
        return aau(fsm, evalArctl(fsm, spec.action),
                   evalArctl(fsm, spec.left),
                   evalArctl(fsm, spec.right))
        
    elif type(spec) is AaW:
        return aaw(fsm, evalArctl(fsm, spec.action),
                   evalArctl(fsm, spec.left),
                   evalArctl(fsm, spec.right))
                   
    elif type(spec) is EaF:
        return eaf(fsm, evalArctl(fsm, spec.action), evalArctl(fsm, spec.child))
        
    elif type(spec) is EaG:
        return eag(fsm, evalArctl(fsm, spec.action), evalArctl(fsm, spec.child))
        
    elif type(spec) is EaX:
        return eax(fsm, evalArctl(fsm, spec.action), evalArctl(fsm, spec.child))
        
    elif type(spec) is EaU:
        return eau(fsm, evalArctl(fsm, spec.action),
                   evalArctl(fsm, spec.left),
                   evalArctl(fsm, spec.right))
        
    elif type(spec) is EaW:
        return eaw(fsm, evalArctl(fsm, spec.action),
                   evalArctl(fsm, spec.left),
                   evalArctl(fsm, spec.right))
                   
    else:
        # TODO Generate error
        print("[ERROR] ARCTL evalArctl: unrecognized specification type", spec)
        return None
        
        
def _fp(funct, start):
    """
    Return the fixpoint of funct, as a BDD, starting with start BDD.
    
    mu Z.f(Z) least fixpoint is implemented with _fp(funct, false).
    nu Z.f(Z) greatest fixpoint is implemented with _fp(funct, true).
    """
    
    old = start
    new = funct(start)
    while old != new:
        old = new
        new = funct(old)
    return old
    
    
def _ex(fsm, alpha, phi):
    """_ex(a, p) is pre of p through transitions satisfying a"""
    return fsm.pre(phi, alpha)
    
    
def _eu(fsm, alpha, phi, psi):
    """_eu(a, p, q) = muZ. (q | (p & _ex(a, Z)))"""
    return _fp(lambda Z: (psi | (phi & _ex(fsm, alpha, Z))),
               BDD.false(fsm.bddEnc.DDmanager))
    
    
def _eg(fsm, alpha, phi):
    """_eg(a, p) = nuZ. (p & _ex(a, Z))"""
    return _fp(lambda Z: (phi & _ex(fsm, alpha, Z)),
               BDD.true(fsm.bddEnc.DDmanager))
    
    
def eax(fsm, alpha, phi):
    """eax(a, p) = _ex(a, p)"""
    return _ex(fsm, alpha, phi)
    
    
def aax(fsm, alpha, phi):
    """aax(a, p) = _ex(a, true) & ~_ex(a, ~p)"""
    return (_ex(fsm, alpha, BDD.true(fsm.bddEnc.DDmanager)) &
          (~_ex(fsm, alpha, (~phi))))
    
    
def eau(fsm, alpha, phi, psi):
    """eau(a, p, q) = _eu(a, p, q)"""
    return _eu(fsm, alpha, phi, psi)
    
    
def aau(fsm, alpha, phi, psi):
    """aau(a, p, q) = ~_eu(a, ~q, ~q & (~p | ~_ex(a, true))) & ~_eg(a, ~q)"""
    return (
                (
                    ~_eu(
                         fsm,
                         alpha,
                         ~psi, 
                         (~psi)
                         &
                         (
                            (~phi)
                            | 
                            (~_ex(fsm, alpha, BDD.true(fsm.bddEnc.DDmanager)))
                         )
                    )
                )
                &
                (
                    ~_eg(fsm, alpha, (~psi))
                )
           )
    

def eaf(fsm, alpha, phi):
    """eaf(a, p) = _eu(a, true, p)"""
    return _eu(fsm, alpha, BDD.true(fsm.bddEnc.DDmanager), phi)
    
    
def aaf(fsm, alpha, phi):
    """aaf(a, p) = ~_eu(a, ~p, ~p & ~_ex(a, true)) & ~_eg(a, ~p)"""
    true = BDD.true(fsm.bddEnc.DDmanager)
    return (
                ~_eu(fsm,
                     alpha,
                     (~phi),
                     (~phi)
                     &
                     (~_ex(fsm, alpha, true))
                    )
                &
                (~_eg(fsm, alpha, (~phi)))
           )
    
    
def eag(fsm, alpha, phi):
    """eag(a, p) = _eu(a, p, p & ~_ex(a, true)) | _eg(a, p)"""
    return _eu(fsm, alpha, phi, (phi &
               (~_ex(fsm, alpha, BDD.true(fsm.bddEnc.DDmanager)))) |
                _eg(fsm, alpha, phi))
    
    
def aag(fsm, alpha, phi):
    """aag(a, p) = ~_eu(a, true, ~p)"""
    return ~_eu(fsm, alpha, BDD.true(fsm.bddEnc.DDmanager), ~phi)

  
def eaw(fsm, alpha, phi, psi):
    """eaw(a, p, q) = ~aau(a, ~q, ~p & ~q)"""
    # TODO Discuss this equivalence with Charles
    return ~aau(fsm, alpha, ~psi, (~phi) & (~psi))
    
    
def aaw(fsm, alpha, phi, psi):    
    # TODO Discuss this equivalence with Charles
    """aaw(a, p, q) = ~eau(a, ~q, ~p & ~q)"""
    return ~eau(fsm, alpha, ~psi, ~phi & ~psi)


import argparse

class NonExitingArgumentParser(argparse.ArgumentParser):
    """An ArgumentParser that does not exit."""
    
    def exit(self, status=0, message=None):
        raise ArgumentParsingError(message)


class ArgumentParsingError(Exception):
    """An error occured while parsing arguments."""
    pass